/*
 * 29/03/2004
 *
 * transferfunction.cpp
 * Thierry Vouriot (Yri)
 * yeri@fr.st
 * http://www.yeri.fr.st
 * thierry.vouriot@free.fr
 * http://thierry.vouriot.free.fr
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "transferfunction.h"

// ============
// Constructeur
// ============
CTransferFunction::CTransferFunction(int i, CImage3D *img) : CWidget(i)
{
	histo = new int[256];

	selected = 0;

	Init(img);
	Maj();
}

void CTransferFunction::Init(CImage3D *img) {

	int ng, i, t;

	cursor1 = 86;
	cursor2 = 170;

	nbVoxels = img->GetWidth() * img->GetHeight() * img->GetDepth();

	for (i=0 ; i<256 ; i++) {
		histo[i] = 0;
	}

	for (i=0; i<nbVoxels; i++) {
		ng = img->GetData()[i];
		histo[ng]++;
	}

	for (i=0; i<NB_CP; i++) {
		cp[i][0] = (int)(i*(255.0f/(float)(NB_CP-1))); 
		cp[i][1] = (int)(i*(128.0f/(float)(NB_CP-1))); 
	}

	// Cration de la liste d'affichage pour l'histogramme
	m_displayList[0] = glGenLists(1);
	glNewList(m_displayList[0], GL_COMPILE);	
	glBegin(GL_LINES);
		glColor3f(1.0f,0.6f,0.0f);
		for (i=0; i<256; i++) {
			if ((t = (int)(((float)(float)histo[i]/(float)(nbVoxels/256.0f))*250)) > 128)
				t = 128;
			glVertex2i(i,10);
			glVertex2i(i,t+10);
		}
	glEnd();
	glEndList();
}

void CTransferFunction::Maj()
{
	int i,k,tmp;
	float ed[NB_CP-1][2];

	for (i = 0; i<NB_CP-1; i++) {
		
		ed[i][0] = (float)(2.0f*(float)cp[i+1][1] - 2.0f*(float)cp[i][1])/(float)((float)cp[i+1][0] - (float)cp[i][0]);
		//printf ("%d %d / %d %d\n", 2.0f*cp[i+1][1],2.0f*cp[i][1],cp[i+1][0],cp[i][0]);
		ed[i][1] = ceilf((((float)2.0f*cp[i+1][1])-(ed[i][0]*(float)cp[i+1][0])));
		//printf("y = %f*x + %f\n", ed[i][0], ed[i][1]);
	}
	
	for (i = 0; i<cursor1*4; i+=4) {
		tf[0+i] = 0;
		tf[1+i] = 0;
		tf[2+i] = 255;
		for (k = 1; k<NB_CP; k++){
			if ((i/4)<=cp[k][0]) break;
		}
		tmp = (int)(((float)(i/4)*(float)ed[k-1][0] + (float)ed[k-1][1]));
		if (tmp > 255) tmp = 255; if (tmp < 0) tmp = 0;
		tf[3+i] = tmp;
		//printf("k=%d  i=%d  a=%d\n",k-1,(i/4),tf[3+i]);
	}

	for (i; i<cursor2*4; i+=4) {
		tf[0+i] = 0;
		tf[1+i] = 255;
		tf[2+i] = 0;
		for (k = 1; k<NB_CP; k++){
			if ((i/4)<=cp[k][0]) break;
		}
		tmp = (int)(((float)(i/4)*(float)ed[k-1][0] + (float)ed[k-1][1]));
		if (tmp > 255) tmp = 255; if (tmp < 0) tmp = 0;
		tf[3+i] = tmp;
		//printf("k=%d  i=%d  a=%d\n",k-1,(i/4),tf[3+i]);
	}

	for (i; i<256*4; i+=4) {
		tf[0+i] = 255;
		tf[1+i] = 0;
		tf[2+i] = 0;
		for (k = 1; k<NB_CP; k++){
			if ((i/4)<=cp[k][0]) break;
		}
		tmp = (int)(((float)(i/4)*(float)ed[k-1][0] + (float)ed[k-1][1]));
		if (tmp > 255) tmp = 255; if (tmp < 0) tmp = 0;
		tf[3+i] = tmp;
		//printf("k=%d  i=%d  a=%d\n",k-1,(i/4),tf[3+i]);
	}	
}

// ===========
// Destructeur
// ===========
CTransferFunction::~CTransferFunction()
{
	delete histo;
	if (glIsList(m_displayList[0])) glDeleteLists(m_displayList[0], 1);
}

void CTransferFunction::Mouse(int button, int state, int mx, int my) {
	if (m_visible) {
	if (button == GLUT_LEFT_BUTTON && state == GLUT_DOWN) {
		mx = mx - x;
		my = my - y;
		selected = RetrieveCursorID(mx, my);
		for(int i=0; i<NB_CP; i++) {
		if (mx > cp[i][0]-3 && mx < cp[i][0]+3 && my > cp[i][1]-3+10 && my < cp[i][1]+3+10)
		selected = i+3;
		}	
	}
	else selected = 0;
	}
}

void CTransferFunction::MouseMotion(int button, int px, int py, int nx, int ny) {
	if (m_visible) {
	if (selected == 1) {
			cursor1 += (nx - px);
			if (cursor1 < 0) cursor1 = 0;
			if (cursor1 >= cursor2) cursor1= cursor2;
		}

		if (selected == 2) {
			cursor2 += (nx - px);
			if (cursor2 >= 255) cursor2 = 256;
			if (cursor2 <= cursor1) cursor2 = cursor1;
		}

		for(int i=0; i<NB_CP; i++) {
			if (selected == i+3) {
			
				if (i != 0 && i != NB_CP-1) {
					cp[i][0] += (nx - px);
					//if (cp[i][0] > 255) cp[i][0] = 255;
					//if (cp[i][0] < 0) cp[i][0] = 0;
					if (cp[i][0] < cp[i-1][0]) cp[i][0] = cp[i-1][0];
					if (cp[i][0] > cp[i+1][0]) cp[i][0] = cp[i+1][0];
				}

			cp[i][1] += (ny - py);
			if (cp[i][1] < 0) cp[i][1] = 0;
			if (cp[i][1] >128) cp[i][1] = 128;
			
			}
		}

		nx = nx - x;
		ny = ny - y;
		if (mouseMotionFunc != NULL)
		if (nx >= 0 && nx <= 255 && ny >= 0 && ny <= 142)
		mouseMotionFunc(id, nx, ny);
	}

}

void CTransferFunction::MousePassiveMotion(int mx, int my) {

	if (m_visible) {
	mx = mx - x;
	my = my - y;
	m_mouseon = RetrieveCursorID(mx, my);
	

	m_cpmouseon = 0;
	for(int i=0; i<NB_CP; i++) {
		if (mx > cp[i][0]-3 && mx < cp[i][0]+3 && my > cp[i][1]-3+10 && my < cp[i][1]+3+10)
		m_cpmouseon = i+1;
	}
	}
}

void CTransferFunction::Draw() {
	
	if (m_visible) {
	int i;

	glPushMatrix();
	glTranslatef((float)x,(float)y,0.0f);
	
	// Dessine l'histogramme
	glCallList(m_displayList[0]);
	
	// Dessine les couleurs sur l'histogramme
	glBegin(GL_QUADS);

	if (m_mouseon == 1) glColor3f(0.1f,0.1f,1.0f);
	else glColor3f(0.0f,0.0f,0.8f);
	glVertex2i(-1,2);
	glVertex2i(cursor1-1,2);
	glVertex2i(cursor1-1,140);
	glVertex2i(-1,140);

	glColor3f(0.0f,0.8f,0.0f);
	glVertex2i(cursor1-1,2);
	glVertex2i(cursor2-1,2);
	glVertex2i(cursor2-1,140);
	glVertex2i(cursor1-1,140);

	if (m_mouseon == 2) glColor3f(1.0f,0.1f,0.1f);
	else glColor3f(0.8f,0.0f,0.0f);
	glVertex2i(cursor2,2);
	glVertex2i(256-1,2);
	glVertex2i(256-1,140);
	glVertex2i(cursor2,140);

	glEnd();

	glBegin(GL_LINES);
	glColor3f(0.8f,0.8f,0.0f);
	glVertex2i(cursor1-1,8);
	glVertex2i(cursor1-1,134);

	glVertex2i(cursor2,8);
	glVertex2i(cursor2,134);
	glEnd();

	glBegin(GL_TRIANGLES);
	
		glVertex2i(cursor1-1,8);
		glVertex2i(cursor1-1-9,0);
		glVertex2i(cursor1-1,0);
	
		glVertex2i(cursor2-1,8);
		glVertex2i(cursor2-1,0);
		glVertex2i(cursor2-1+9,0);

		glVertex2i(cursor1-1,142);
		glVertex2i(cursor1-1-9,142);
		glVertex2i(cursor1-1,134);

		glVertex2i(cursor2-1,142);		
		glVertex2i(cursor2-1,134);
		glVertex2i(cursor2-1+9,142);
	glEnd();

	glDisable(GL_BLEND);

	// Dessine la courbe et les points de contrle
	glColor3f(0.9f,0.9f,0.9f);
	glBegin(GL_LINES);
	for (i = 1; i<NB_CP; i++) {
		glVertex2i(cp[i-1][0],cp[i-1][1]+10);
		glVertex2i(cp[i][0],cp[i][1]+10);
	}
	glEnd();

	glBegin(GL_QUADS);
	for (i = 0; i<NB_CP; i++) {
		if (m_cpmouseon == i+1) glColor3f(0.4f,0.4f,0.4f);
		else glColor3f(0.9f,0.9f,0.9f);
		glVertex2i(cp[i][0]+3,cp[i][1]-3+10);
		glVertex2i(cp[i][0]+3,cp[i][1]+3+10);
		glVertex2i(cp[i][0]-3,cp[i][1]+3+10);
		glVertex2i(cp[i][0]-3,cp[i][1]-3+10);
	}
	glEnd();

	glEnable(GL_BLEND);

	glPopMatrix();
	}
}

int CTransferFunction::RetrieveCursorID(int mx, int my)
{
	int k;

	if ((9 - cursor1) >= 0) k = -(9-cursor1); else k = 0;
	if (mx <= cursor1-1 && mx >= k)
		if (my <= 139 && my >=0)
			return 1;

	if ((255-cursor2) <= 9) k = 255+(9-(255-cursor2)); else k = 255;
	if (mx <= k && mx >= cursor2-1)
		if (my <= 139 && my >=0)
			return 2;

	return 0;
}


