/*
 * 29/03/2004
 *
 * main.cpp	(CAVELIB)
 * Thierry Vouriot (Yri)
 * yeri@fr.st
 * http://www.yeri.fr.st
 * thierry.vouriot@free.fr
 * http://thierry.vouriot.free.fr
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "main.h"

CTimerGLUT *g_timer;
GLuint g_textures[1];
GLfloat viewDistance = -2.5;

int slices = 80, lastslices;
float intensity = 0.30;
float density = 0.1;
float scale[3] = {1.0, 1.0, 1.0};
CImage3D *i3d;
CMyGui *mygui;
GLenum filter = GL_LINEAR, lastfilter;
bool aff = true;

#define ZOOM 0
#define TFRED 2
#define TFBLUE 3
#define BLENDING 4
#define ROTATION 1
#define NBMODES 5
int mode = ZOOM;
int blending = 0;

float lastPos[3];
GLfloat rotX = 0.0f, rotY = 0.0f;

int w,h,d,t;
char file[1024];

// ====
// Main
// ====
int main(int argc, char ** argv)
{	
	/* Initialize the CAVE */
	CAVEConfigure(&argc,argv,NULL);

	if (argc != 6 && argc != 1) { printf("Usage: visuvo filename width height depth [8|16]\n"); exit(0); }

	// Volume  charger avec ses paramtres
	if (argc == 1) {
		w = h = d = 256;
		t = true;
		strcpy(file,"data/foot.raw");
	}
	else {
		w = atoi(argv[2]);
		h = atoi(argv[3]);
		d = atoi(argv[4]);
		t = (8 == atoi(argv[5]));
		strcpy(file,argv[1]);
	}


	/* Give the library a pointer to the GL initialization function */
	CAVEInitApplication(initOpenGL,0);

	/* Give the library a pointer to the drawing function */
	CAVEDisplay(caveDisplay,0);

	/* Give the library a pointer to an update function */
	CAVEFrameFunction((CAVECALLBACK)caveIdle,0);

	/* Create the multiple processes/threads and start the display loop */
	CAVEInit();

	/* Wait for the escape key to be hit */
	while (!CAVEgetbutton(CAVE_ESCKEY))
	{
		/* Nap so that this busy loop doesn't waste CPU time reset timeval
		struct every time for linux compatibility
		*/
		sleep(1);
	}

	/* Clean up & exit */
	CAVEExit();
	return 0;
}

// ========================
// Initialisation de OpenGL
// ========================
void initOpenGL(void)
{
	/* Affiche des informations sur OpenGL */
	char *exten;
	exten = (char *) glGetString(GL_VENDOR);
	printf("%s\n",exten);
	exten = (char *) glGetString(GL_RENDERER);
	printf("%s\n",exten);
	exten = (char *) glGetString(GL_VERSION);
	printf("%s\n",exten);

	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
	glClearDepth(1.);
	glDepthFunc (GL_LEQUAL);									// The Type Of Depth Testing (Less Or Equal)
	glEnable (GL_DEPTH_TEST);									// Enable Depth Testing
	glEnable(GL_CULL_FACE);										// permet de masquer les faces arrires	
	glShadeModel (GL_SMOOTH);									// Select Smooth Shading
	glHint (GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);			// Set Perspective Calculations To Most Accurate
	glBlendFunc(GL_SRC_ALPHA,GL_ONE);							// Enable Blending

	init();
}

// ========================
// Initialisations diverses
// ========================
void init()
{
	// Allocation des variables globales
	g_timer = new CTimerGLUT();

	// Extensions OpenGL disponibles
	initExtensions();
	CString missing;
	if (!GL_EXT_texture3D_supported)					missing += "GL_EXT_texture3D\n";

    if (missing.getLength() > 0){
		printf("\nThe following required OpenGL extensions isn't supported by the graphic card and driver you're using:\n\n" + missing +
                   "\nThis application will now exit.");
		#ifdef _WIN32
		MessageBox(NULL,"The following required OpenGL extensions isn't supported by the graphic card and driver you're using:\n\n" + missing +
                   "\nThis application will now exit.\n", "Missing extensions", MB_OK);
		#endif
		exit(-1);
    }

	// Chargement du volume
	i3d = new CImage3D();
	i3d->LoadRAW(file, w, h, d, t);
	
	scale[0] =  i3d->GetScale()[0]/2;
	scale[1] =  i3d->GetScale()[1]/2;
	scale[2] =  i3d->GetScale()[2]/2;	

	// Initialisation de l'interface graphique
	mygui = new CMyGui(0, i3d);
	mygui->SetPosition(10, 10);
	mygui->SetShow(MENU);
	mygui->size->SetLabel("SIZE: %dx%dx%d", i3d->GetWidth(), i3d->GetHeight(), i3d->GetDepth());
	
	// Cration de la texture 3D
	if (GL_EXT_paletted_texture_supported) i3d->CreateTexture_TF_PalettedTexture(g_textures, 0, mygui->tf->tf);
	else i3d->CreateTexture_TF(g_textures, 0, mygui->tf->tf);

}

// ================================
// Fonction d'affichage de la scene
// ================================
void caveDisplay(void)
{
	int i;
	
	// TexGen planes
	float xPlane[]={ 1.0f, 0.0f, 0.0f, 0.0f };
	float yPlane[]={ 0.0f, 1.0f, 0.0f, 0.0f };
	float zPlane[]={ 0.0f, 0.0f, 1.0f, 0.0f };
	
	glMatrixMode(GL_MODELVIEW);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glPushMatrix();

	CAVENavTransform();

	// On se dplace  une certaine distance
	glTranslatef(0.0f,1.0f,viewDistance);
	
	// Pour la gnration des coordonnes des textures
  	glTexGeni(GL_S,GL_TEXTURE_GEN_MODE,GL_EYE_LINEAR);
  	glTexGeni(GL_T,GL_TEXTURE_GEN_MODE,GL_EYE_LINEAR);
  	glTexGeni(GL_R,GL_TEXTURE_GEN_MODE,GL_EYE_LINEAR);  
	
	// On passe  la matrice de texture
	glMatrixMode(GL_TEXTURE);
	glEnable(GL_TEXTURE_3D_EXT);
	glLoadIdentity();
	
	// Applique les transformations  l'objet
	glTranslatef(0.5f, 0.5f, 0.5f);
  	glScalef(scale[0],scale[1],scale[2]);
	
	glRotatef(-rotX, 1.0f, 0.0f, 0.0f);
	glRotatef(-rotY, 0.0f, 1.0f, 0.0f);
	
	// On applique la translation inverse pour la distance de vue
  	glTranslatef(0.0f,-1.0f,-viewDistance);	
	
	// On retourne  la matrice MODELVIEW
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
	glPushMatrix();
	
	// Dfinition des plans pour TexGen
  	glTexGenfv(GL_S,GL_EYE_PLANE,xPlane);
  	glTexGenfv(GL_T,GL_EYE_PLANE,yPlane);
  	glTexGenfv(GL_R,GL_EYE_PLANE,zPlane);

	// Active la gnration automatique des coordonnes de textures
 	glEnable(GL_TEXTURE_GEN_S);
  	glEnable(GL_TEXTURE_GEN_T);
  	glEnable(GL_TEXTURE_GEN_R);
	
	// Initialise la couleur des tranches du volume
	glColor4f(1.0,1.0,1.0, intensity);
	glEnable(GL_TEXTURE_3D_EXT);
	glBindTexture(GL_TEXTURE_3D_EXT, g_textures[0]);
	
	// Filtrage des textures
	glTexParameteri(GL_TEXTURE_3D_EXT,GL_TEXTURE_MAG_FILTER,filter);
	glTexParameteri(GL_TEXTURE_3D_EXT,GL_TEXTURE_MIN_FILTER,filter);
	
	// Active le test alpha blending
	
	glEnable(GL_ALPHA_TEST);

	if (false) {
		glAlphaFunc(GL_GREATER, density*intensity);
	}
	else glAlphaFunc(GL_GREATER, 0);

	// Choix de la fonction de blending
	if (GL_EXT_blend_minmax_supported) glBlendEquationEXT(GL_FUNC_ADD_EXT);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE);
	if (blending == 0) {
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	}
	else if (blending == 2) {
		glBlendFunc(GL_CONSTANT_ALPHA_EXT, GL_ONE);
		glBlendColorEXT(1.f, 1.f, 1.f, 1.f/slices*3.0f);
	}
	else if (blending == 3) {
		glBlendEquationEXT(GL_MAX_EXT);
	}
	glEnable(GL_BLEND);
	
	// On se dplace  une certaine distance
	glTranslatef(0.0f,1.0f,viewDistance);
	
	// Dessine les tranches du volumes
	for (i=0; i < slices; i++)
	{
		glPushMatrix();
		// Z-translation pour dessiner la tranche courante
		glTranslatef(0.0, 0.0, -1.0 + (float)(i)*2.0f/(float)(slices-1));
		
		// Dessin de la tranche
		// Les coordonnes de textures sont assignes automatiquement
		// grace  l'utilisation de TexGen
		glBegin(GL_QUADS);
			glVertex3f(-1.0, -1.0, 0.0);
			glVertex3f(1.0, -1.0, 0.0);
			glVertex3f(1.0, 1.0, 0.0);
			glVertex3f(-1.0, 1.0, 0.0);
		glEnd();
		glPopMatrix();
	}

	glDisable(GL_ALPHA_TEST);
	glDisable(GL_BLEND);
	glDisable(GL_TEXTURE_3D_EXT);
	glPopMatrix();
	
	// Desactive la gnration automatique des coordonnes de textures
 	glDisable(GL_TEXTURE_GEN_S);
  	glDisable(GL_TEXTURE_GEN_T);
  	glDisable(GL_TEXTURE_GEN_R);

	
	if (aff) {
		mygui->SetVisible(true);
		// Dessine un cube autour de la texture
		if (mode == ZOOM) glColor4f(0.0f,0.7f,0.0f,1.0f);
		else if (mode == TFRED) glColor4f(0.7f,0.0f,0.0f,1.0f);
		else if (mode == TFBLUE) glColor4f(0.0f,0.0f,0.7f,1.0f);
		else if (mode == BLENDING) glColor4f(0.0f,0.7f,0.7f,1.0f);
		else if (mode == ROTATION) glColor4f(0.7f,0.7f,0.0f,1.0f);
		
		//glutWireCube(2.0);

		// Dessine l'interface utilisateur
		glBlendFunc(GL_SRC_ALPHA, GL_ONE);
		glEnable(GL_BLEND);
	
		mygui->fps->SetLabel("%4d fps", g_timer->GetFPS());
		if (mode == ZOOM) mygui->mode->SetLabel("MODE: ZOOM");
		if (mode == ROTATION) mygui->mode->SetLabel("MODE: ROTATION");
		if (mode == TFRED) mygui->mode->SetLabel("MODE: TRANSFER_FUNCTION_RED");
		if (mode == TFBLUE) mygui->mode->SetLabel("MODE: TRANSFER_FUNCTION_BLUE");
		if (mode == BLENDING) mygui->mode->SetLabel("MODE: BLENDING %d",blending);
		mygui->Draw(1375,900);
	}
	else mygui->SetVisible(false);

}

// ===========================================================
// Fonction excut lorsque le programme n'a plus rien  faire
// utilisee pour les animations
// ===========================================================
void caveIdle(void)
{
    float jx, jy, dt, t;
    static float prevtime = 0;
    static bool prevButton1 = false;
    static bool prevButton2 = false;
    
    static bool prevJoyX = false;
    static bool prevJoyY = false;

    if(CAVEMasterDisplay()) {
	
		// Calcul le framerate
		g_timer->CalcFPS();

        	jy=CAVE_JOYSTICK_X;
        	jx=CAVE_JOYSTICK_Y;
        	t = (float) CAVEGetTime();
        	dt = t - prevtime;

        	prevtime = t;

		if (CAVEBUTTON2 && !prevButton2)  { mode++; prevButton2 = true;} else if (!CAVEBUTTON2) prevButton2 = false;
		if (CAVEBUTTON1 && !prevButton1)  { mode--; prevButton1 = true;} else if (!CAVEBUTTON1) prevButton1 = false;
		if (mode >= NBMODES) mode = 0;
		if (mode < 0) mode = NBMODES-1;

		if (mode == ROTATION) {
			if (fabs(jy)>0.1) {
            		//float wandFront[3];
            		//CAVEGetVector(CAVE_WAND_FRONT,wandFront);
            		//CAVENavTranslate(wandFront[0]*jy*SPEED*dt, wandFront[1]*jy*SPEED*dt,wandFront[2]*jy*SPEED*dt);
				rotY += jy*dt*90;
				if (rotY > 360.0f) rotY = 0.0f;
				if (rotY < -360.0f) rotY = 0.0f;
        		}
        		if (fabs(jx)>0.1) {
            		//CAVENavRot(-jx*90.0f*dt,'y');
				rotX += jx*dt*90;
				if (rotX > 360.0f) rotX = 0.0f;
				if (rotX < -360.0f) rotX = 0.0f;
        		}
		}
		else if (mode == ZOOM) {
			if (fabs(jx)>=0.8 && jx > 0.0)  viewDistance += 0.2;
			if (fabs(jx)>=0.8 && jx < 0.0)  viewDistance -= 0.2;
			if (viewDistance < -10.0) viewDistance = -10.0;
			if (viewDistance > 0.0) viewDistance = 0.0;
		}
		else if (mode == BLENDING) {
			if (fabs(jy)>=1.0 && jy > 0.0 && !prevJoyY)  { blending++; prevJoyY = true;} else if (jy == 0.0) prevJoyY = false;
			if (fabs(jy)>=1.0 && jy < 0.0 && !prevJoyY)  { blending--; prevJoyY = true;} else if (jy == 0.0) prevJoyY = false;
			if (blending >= 4) blending = 0;
			if (blending < 0) blending = 3;
		}
		else if (mode == TFRED) {
			if (fabs(jy)>=0.8 && jy > 0.0)  mygui->tf->cursor2++;
			if (fabs(jy)>=0.8 && jy < 0.0)  mygui->tf->cursor2--;
			
			if (mygui->tf->cursor2 >= 256) mygui->tf->cursor2 = 256;
			if (mygui->tf->cursor1 >= mygui->tf->cursor2) mygui->tf->cursor2 = mygui->tf->cursor1;
			if (GL_EXT_paletted_texture_supported) {
				mygui->tf->Maj();
				glBindTexture(GL_TEXTURE_3D_EXT, g_textures[0]);
				glColorTableEXT(GL_TEXTURE_3D_EXT,GL_RGBA8,256,GL_RGBA,GL_UNSIGNED_BYTE,mygui->tf->tf);
			}
		}
		else if (mode == TFBLUE) {
			if (fabs(jy)>=0.8 && jy > 0.0)  mygui->tf->cursor1++;
			if (fabs(jy)>=0.8 && jy < 0.0)  mygui->tf->cursor1--;

			if (mygui->tf->cursor1 <= 0) mygui->tf->cursor1 = 0;
			if (mygui->tf->cursor2 <= mygui->tf->cursor1) mygui->tf->cursor1 = mygui->tf->cursor2;
			if (GL_EXT_paletted_texture_supported) {
				mygui->tf->Maj();
				glBindTexture(GL_TEXTURE_3D_EXT, g_textures[0]);
				glColorTableEXT(GL_TEXTURE_3D_EXT,GL_RGBA8,256,GL_RGBA,GL_UNSIGNED_BYTE,mygui->tf->tf);
			}
		}
    }
	CAVEDisplayBarrier();
}

