 /*
 * 29/03/2004
 *
 * string.cpp
 * Thierry Vouriot (Yri)
 * yeri@fr.st
 * http://www.yeri.fr.st
 * thierry.vouriot@free.fr
 * http://thierry.vouriot.free.fr
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "string.h"
#include <stdarg.h>
#include <stdio.h>

int getAppropriateCapasity(unsigned int capasity, unsigned int min){
	do {
		capasity <<= 1;
	} while (capasity < min);
	return capasity;
}


CString::CString(){
	length = 0;
	str = (char *) malloc(capasity = 32);
	str[0] = '\0';
}

CString::CString(unsigned int cap){
	length = 0;
	str = (char *) malloc(capasity = cap);
	str[0] = '\0';
}

CString::CString(const char *string){
	length = strlen(string);
	str = (char *) malloc(capasity = length + 1);
	strncpy(str, string, capasity);
}

CString::CString(const char *string, unsigned int len){
	length = len;
	str = (char *) malloc(capasity = len + 1);
	strncpy(str, string, len);
	str[len] = '\0';
}

CString::CString(const CString &string){
	length = string.length;
	capasity = string.capasity;
	str = (char *) malloc(capasity);
	strncpy(str, string.str, length + 1);
}


CString::~CString(){
	free(str);
}

void CString::setLength(const unsigned int newLength){
	str = (char *) realloc(str, capasity = newLength + 1);
	str[length = newLength] = '\0';
}


void CString::makeLowerCase(){
	for (unsigned int i = 0; i < length; i++){
		if (str[i] >= 'A' && str[i] <= 'Z') str[i] -= ('A' - 'a');
	}
}


void CString::operator = (const char *string){
	assign(string, strlen(string));
}

void CString::operator = (const CString &string){
	assign(string.str, string.length);
}

void CString::operator += (const char *string){
	append(string, strlen(string));
}

void CString::operator += (const CString &string){
	append(string.str, string.length);
}

void CString::assign(const char *string, const unsigned int len){
	if (len + 1 >= capasity){
		capasity = getAppropriateCapasity(capasity, len + 1);
		free(str);
		str = (char *) malloc(capasity);
	}

	strncpy(str, string, len);
	str[len] = '\0';
	length = len;
}

void CString::append(const char *string, const unsigned int len){
	unsigned int newLength = length + len;
	if (newLength + 1 >= capasity){
		capasity = getAppropriateCapasity(capasity, newLength + 1);
		str = (char *) realloc(str, capasity);
	}

	strncpy(str + length, string, len);
	str[newLength] = '\0';

	length = newLength;
}

bool CString::insert(const unsigned int pos, const char *string, const unsigned int len){
	if (pos > length) return false;

	unsigned int newLength = length + len;
	if (newLength + 1 >= capasity){
		capasity = getAppropriateCapasity(capasity, newLength + 1);
		str = (char *) realloc(str, capasity);
	}

	unsigned int n = length - pos;
	for (unsigned int i = 0; i <= n; i++){
		str[newLength - i] = str[length - i];
	}
	strncpy(str + pos, string, len);

	length = newLength;
	return true;
}

bool CString::remove(const unsigned int pos, const unsigned int len){
	if (pos + len > length) return false;

	length -= len;
	unsigned int n = length - pos;
	for (unsigned int i = 0; i <= n; i++){
		str[pos + i] = str[pos + len + i];
	}

	return true;
}

unsigned int CString::replace(const char oldCh, const char newCh){
	unsigned int count = 0;
	for (unsigned int i = 0; i < length; i++){
		if (str[i] == oldCh){
			str[i] = newCh;
			count++;			
		}
	}
	return count;
}

unsigned int CString::replace(const char *oldStr, const char *newStr){
	unsigned int i, j, last, count = 0;
	const unsigned int oldLen = strlen(oldStr);
	const unsigned int newLen = strlen(newStr);

	for (i = 0; i < length - oldLen; i++){
		if (strncmp(str + i, oldStr, oldLen) == 0) count++;
	}

	length += (newLen - oldLen) * count;
	capasity = getAppropriateCapasity(capasity, length + 1);

	char *dest = (newLen > oldLen)? (char *) malloc(capasity) : str;
	char *newString = dest;

	i = 0;
	last = 0;
	for (j = 0; j < count; j++){
		while (strncmp(str + i, oldStr, oldLen) != 0) i++;

		if (j || newLen > oldLen) strncpy(dest, str + last, i - last);
		dest += i - last;
		strncpy(dest, newStr, newLen);
		dest += newLen;
		i += oldLen;
		last = i;
	}

	strcpy(dest, str + i);

	if (newLen > oldLen){
		free(str);
		str = newString;
	}

	return count;
}

bool CString::find(const char ch, unsigned int pos, unsigned int *index) const {
	for (unsigned int i = pos; i < length; i++){
		if (str[i] == ch){
			if (index != NULL) *index = i;
			return true;
		}
	}
	return false;
}

bool CString::rfind(const char ch, int pos, unsigned int *index) const {
	unsigned int i = (pos < 0)? length : pos;

	while (i){
		i--;
		if (str[i] == ch){
			if (index != NULL) *index = i;
			return true;
		}
	}
	return false;
}

bool CString::find(const char *string, unsigned int pos, unsigned int *index) const {
	char *st = strstr(str + pos, string);
	if (st != NULL){
		if (index != NULL) *index = st - str;
		return true;
	}
	return false;
}


bool isInString(char *str, const char ch){
	while (*str){
		if (*str++ == ch) return true;
	}
	return false;
}

void CString::trimRight(const char *chars){
	int i = length - 1;
	while (i >= 0 && isInString((char *) chars, str[i])) i--;

	length = i + 1;
	str[length] = '\0';
}

void CString::sprintf(char *formatStr, ... ){
	char buf[256];
	union {
		double farg;
		int    iarg;
		char  *sarg;
	};
	va_list marker;
	va_start(marker, formatStr);

	int len, i = 0;
	int last = 0;
	int flen = strlen(formatStr);
	do {
		while (i < flen && formatStr[i] != '%') i++;
		append(formatStr + last, i - last);
		if (i >= flen){
			va_end(marker);
			return;
		}

		switch (formatStr[i + 1]){
		case 'd':
			iarg = va_arg(marker, int);
			len = ::sprintf(buf, "%d", iarg);
			append(buf, len);
			break;
		case 'f':
			farg = va_arg(marker, double);
			len = ::sprintf(buf, "%f", farg);
			append(buf, len);
			break;
		case 'g':
			farg = va_arg(marker, double);
			len = ::sprintf(buf, "%g", farg);
			append(buf, len);
			break;
		case 's':
			sarg = va_arg(marker, char *);
			operator += (sarg);
			break;
		case 'x':
			iarg = va_arg(marker, int);
			len = ::sprintf(buf, "%x", iarg);
			append(buf, len);
			break;
		case 'X':
			iarg = va_arg(marker, int);
			len = ::sprintf(buf, "%X", iarg);
			append(buf, len);
			break;
		case '%':
			append("%", 1);
			break;
		}
		i += 2;
		last = i;
	} while (true);
}


CString operator + (const CString &string, const CString &string2){
	CString str(string.getLength() + string2.getLength() + 1);
	str = string;
	str.append(string2, string2.getLength());
	return str;
}

CString operator + (const CString &string, const char *string2){
	int len = strlen(string2);
	CString str(string.getLength() + len + 1);
	str = string;
	str.append(string2, len);
	return str;
}

CString operator + (const char *string, const CString &string2){
	int len = strlen(string);
	CString str(string2.getLength() + len + 1);
	str = string;
	str.append(string2, string2.getLength());
	return str;
}

// Quick implementation macro for all comparison functions
#define IMP_COMPARE(s0,op,s1) bool operator op (const s0 string, const s1 string2){ return (strcmp(string, string2) op 0); }

IMP_COMPARE(CString &, ==, CString &)
IMP_COMPARE(CString &, ==, char *)
IMP_COMPARE(char *  , ==, CString &)

IMP_COMPARE(CString &, !=, CString &)
IMP_COMPARE(CString &, !=, char *)
IMP_COMPARE(char *  , !=, CString &)

IMP_COMPARE(CString &, >, CString &)
IMP_COMPARE(CString &, >, char *)
IMP_COMPARE(char *  , >, CString &)

IMP_COMPARE(CString &, <, CString &)
IMP_COMPARE(CString &, <, char *)
IMP_COMPARE(char *  , <, CString &)

IMP_COMPARE(CString &, >=, CString &)
IMP_COMPARE(CString &, >=, char *)
IMP_COMPARE(char *  , >=, CString &)

IMP_COMPARE(CString &, <=, CString &)
IMP_COMPARE(CString &, <=, char *)
IMP_COMPARE(char *  , <=, CString &)
