/*
 * 29/03/2004
 *
 * string.h
 * Classe String
 *
 * Author: Thierry Vouriot (Yri)
 * yeri@fr.st
 * http://www.yeri.fr.st
 * thierry.vouriot@free.fr
 * http://thierry.vouriot.free.fr
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef _STRING_H_
#define _STRING_H_

#include <string.h>
#include <stdlib.h>

class CString {
private:
	char *str;
	unsigned int length;
	unsigned int capasity;
public:
	CString();
	CString(unsigned int cap);
	CString(const char *string);
	CString(const char *string, unsigned int len);
	CString(const CString &string);
	~CString();
	
	void setLength(const unsigned int newLength);

	operator const char *() const { return str; }
	unsigned int getLength() const { return length; }
	bool isEmpty() const { return length == 0; }

	int toInt() const { return atoi(str); }
	float toFloat() const { return (float) atof(str); }
	double toDouble() const { return atof(str); }

	void makeLowerCase();

	void operator =  (const char *string);
	void operator =  (const CString &string);
	void operator += (const char *string);
	void operator += (const CString &string);

	void assign(const char *string, const unsigned int len);
	void append(const char *string, const unsigned int len);
	bool insert(const unsigned int pos, const char *string, const unsigned int len);
	bool remove(const unsigned int pos, const unsigned int len);

	unsigned int replace(const char oldCh, const char newCh);
	unsigned int replace(const char *oldStr, const char *newStr);

	bool  find(const char ch, unsigned int pos = 0, unsigned int *index = NULL) const;
	bool rfind(const char ch, int pos = -1, unsigned int *index = NULL) const;
	bool  find(const char *string, unsigned int pos = 0, unsigned int *index = NULL) const;

	void trimRight(const char *chars);

	void sprintf(char *formatStr, ... );
};

CString operator + (const CString &string, const CString &string2);
CString operator + (const CString &string, const char   *string2);
CString operator + (const char   *string, const CString &string2);

bool operator == (const CString &string, const CString &string2);
bool operator == (const CString &string, const char   *string2);
bool operator == (const char   *string, const CString &string2);

bool operator != (const CString &string, const CString &string2);
bool operator != (const CString &string, const char   *string2);
bool operator != (const char   *string, const CString &string2);

bool operator > (const CString &string, const CString &string2);
bool operator > (const CString &string, const char   *string2);
bool operator > (const char   *string, const CString &string2);

bool operator < (const CString &string, const CString &string2);
bool operator < (const CString &string, const char   *string2);
bool operator < (const char   *string, const CString &string2);

bool operator >= (const CString &string, const CString &string2);
bool operator >= (const CString &string, const char   *string2);
bool operator >= (const char   *string, const CString &string2);

bool operator <= (const CString &string, const CString &string2);
bool operator <= (const CString &string, const char   *string2);
bool operator <= (const char   *string, const CString &string2);

#endif // _STRING_H_
