/*
 * 05/28/2003
 *
 * timerGLUT.cpp
 * Author: Thierry Vouriot (Yri)
 * yeri@fr.st
 * http://www.yeri.fr.st
 * thierry.vouriot@free.fr
 * http://thierry.vouriot.free.fr
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "timerGLUT.h"

// #################
// Classe CTimerGLUT
// #################


// =======================
// Constructeur par defaut
// =======================
CTimerGLUT::CTimerGLUT() 
{ 
	m_startTime = glutGet(GLUT_ELAPSED_TIME);
	m_frameInterval = 0;
	m_fps = 0;
}


// ===========
// Destructeur
// ===========
CTimerGLUT::~CTimerGLUT() { }


// ==============================================================
// Retourne le temps en millisecondes entre 2 frames
// Attention cette valeur n'est disponible que lorsque
// GetFPS() ou GetElapsedSeconds() on ete utilise
// mais on ne doit utiliser qu'une seule de ces 2 fonctions
// entre l'affichage de 2 frames sinon celle-ci n'est plus valide
// ==============================================================
int CTimerGLUT::FrameInterval() 
{ 
	return m_frameInterval;
}


// ==============================================================
// Retourne le temps en millisecondes depuis la creation du timer
// ou depuis le dernier appel de GetElapsedSeconds()
// ==============================================================
int CTimerGLUT::GetElapsedSeconds()
{
	static int s_lastTime = m_startTime;
	int currentTime = glutGet(GLUT_ELAPSED_TIME);

	int seconds = currentTime - s_lastTime;

	m_frameInterval = currentTime - s_lastTime;
	s_lastTime = currentTime;
		
	return seconds * 1000;
}


// ======================================================
// Calcul le nombre de frames par secondes de l'affichage
// ======================================================
int CTimerGLUT::CalcFPS()
{
	static int s_timeAtLastFrame = m_startTime;	
	int currentTime = glutGet(GLUT_ELAPSED_TIME);
	m_frameInterval = currentTime - s_timeAtLastFrame; // Temps coul entre l'affichage de 2 images
	

	static int fps = 0;					// frames per second	
	static float s_lastTime = 0.0f;		// temps au moment du dernier affichage du framerate	
	
  	++fps;
	
	// si une seconde est passe
	if( s_timeAtLastFrame - s_lastTime > 1000.0f )
  	{
		s_lastTime = s_timeAtLastFrame;
		m_fps = fps;
		fps = 0;	// rinitialise le nombre d'images

	}

	s_timeAtLastFrame = currentTime;	
	return m_fps;	
}


// ======================================
// Retourne le dernier frame rate calcule
// ======================================
int CTimerGLUT::GetFPS()
{
	return m_fps;
}
